/* 
 * Copyright (C) 2008 Laird Breyer
 *  
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * Author:   Laird Breyer <laird@lbreyer.com>
 */

#include "imagebuffer.h"
#include "bittwiddler.h"
#include "mask.h"
#include "tessboxes.h"

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

extern options_t u_options;

bool_t create_imagebuffer(imagebuffer_t *ib) {
  if( ib ) {
    ib->image = NULL;
    ib->image_size = 0;
    return clear_imagebuffer(ib);
  }
  return FALSE;
}

bool_t clear_imagebuffer(imagebuffer_t *ib) {
  if( ib ) {
    ib->width = 0;
    ib->height = 0;
    ib->pos.x = 0;
    ib->pos.y = 0;

    memset(&ib->colours, 0, sizeof(ib->colours));
    ib->colours.min = 0;
    ib->colours.max = MAX_GS;

    ib->crop.xmin = ib->width;
    ib->crop.xmax = 0;
    ib->crop.ymin = ib->height;
    ib->crop.ymax = 0;
    ib->crop.coverage = 0;
    return TRUE;
  }
  return FALSE;
}


/* grow the image, but preserve its contents */
bool_t grow_imagebuffer(imagebuffer_t *ib, int w, int h) {
  byte_t *newmem;
  size_t oldwidth, oldrowstart, newrowstart, newsize;
  int i;

  if( ib ) {
    w = MAX(w,ib->width);
    h = MAX(h,ib->height);
    if( (w + h) > (ib->width + ib->height) ) {
      newsize = w * h * sizeof(imagepoint_t);
      if( newsize > ib->image_size ) {
	newmem = realloc(ib->image, newsize);
	if( !newmem ) {
	  return FALSE;
	}
	ib->image = (imagepoint_t *)newmem;
	ib->image_size = newsize;
      }

      /* rearrange the data, and fill unused space with zeros */
      /* work backwards from last row, and use memmove not memcpy */

      for(i = ib->height - 1; i >= 0; i--) {
	newrowstart = i * w;
	oldrowstart = i * ib->width;
	oldwidth = ib->width;
	memmove(ib->image + newrowstart, ib->image + oldrowstart, 
		oldwidth * sizeof(imagepoint_t));
	memset(ib->image + newrowstart + oldwidth, 0,
	       (w - oldwidth) * sizeof(imagepoint_t));
      }

      /* all done, clean up */
      ib->width = w;
      ib->height = h;
    }
    return TRUE;
  }
  return FALSE;
}

bool_t free_imagebuffer(imagebuffer_t *ib) {
  if( ib ) {
    if( ib->image_size > 0 ) {
      free(ib->image);
      ib->image = NULL;
      ib->image_size = 0;
      return TRUE;
    }
  }
  return FALSE;
}

bool_t seekpos_imagebuffer(imagebuffer_t *ib, int x, int y) {
  if( ib && (x <= ib->width) && (y <= ib->height) ) {
    ib->pos.x = x;
    ib->pos.y = y;
    return TRUE;
  }
  return FALSE;
}



/* This draws (a piece of) an image of size (width,height) at the top
 * left corner of the imagebuffer (pixl0). The piece is given by a bit
 * array in buffer, and (ib->pos.x,ib->pos.y) keeps track of where we
 * write the piece.
 * Note that we rely on the fact that width <= ib->width, height <= ib->height,
 * but this is guaranteed if you've called grow_imagebuffer() when reading
 * each image.
 */
bool_t partfill_bw_imagebuffer(imagebuffer_t *ib, byte_t *buffer, 
			       int numbits, int width, int height,
			       int plane) {
  bittwiddler_t bt;
  unsigned int x, y, pixel;
  imagepoint_t *point;

  if( ib && buffer && (numbits > 0) ) {
    x = ib->pos.x;
    y = ib->pos.y;
    if( x >= width ) { 
      x = 0; 
      y++; 
    }
    while(y < height) {
      point = ib->image + (y * ib->width) + x;
      setbyte_bittwiddler(&bt, *buffer);
      rewind_bittwiddler(&bt);
      while(x < width) {
	if( readnextbit_bittwiddler(&bt, &pixel) ) {
	  WRITEBW(point->pixl,plane,pixel);
	  point++;
	  x++;
	  if( --numbits <= 0) {
	    goto done;
	  }
	} else {
	  buffer++;
	  setbyte_bittwiddler(&bt, *buffer);
	}
      }
      x = 0;
      y++;
      /* last byte in row is padded with garbage */
      if( flush_bittwiddler(&bt) ) {
	buffer++;
	setbyte_bittwiddler(&bt, *buffer);
      }
    }
  done:
    if( numbits > 0 ) {
      /* this is serious */
      return FALSE;
    }
    /* update write position */
    seekpos_imagebuffer(ib, x, y);
  }
  return FALSE;
}

void render_pixel_imagebuffer(imagepoint_t *point, int plane,
			      byte_t *red, byte_t *green, byte_t *blue) {
  /* default colour white */
  *red = 128;
  *green = 128;
  *blue = 128;

  /* underlying image */  
  if( READBW(point->pixl, 0) ) {
    *red = 0;
    *green = 0;
    *blue = 0;
  }
  
  if( u_options & FL_BB ) {
    /* bounding boxes */    
    if( READBW(point->pixl, 1) ) {
      *green = 255;
      *red = 0;
      *blue = 0;
    }
  }
}

bool_t save_color_imagebuffer(imagebuffer_t *ib, ppmwriter_t *p, 
			      imagetype_t type, int plane) {
  byte_t *buf, *writepos;
  size_t bufsize;
  imagepoint_t *point;
  int x, y;
  bool_t retval = TRUE;
  int xmin, xmax, ymin, ymax, c;
  byte_t red, green, blue;

  if( ib && p && p->f ) {

    xmin = 0;
    xmax = ib->width;
    ymin = 0;
    ymax = ib->height;

    c = 16;

    if( write_header_ppmwriter(p, xmax - xmin, ymax - ymin, c, tP6) ) {
      bufsize = 3 * (ib->width * sizeof(byte_t) + 1); /* one full row, KISS */
      buf = malloc(bufsize * sizeof(byte_t));
      if( !buf ) {
	retval = FALSE;
      } else {
	for(y = ymin; y < ymax; y++) {
	  point = ib->image + y * ib->width + xmin;
	  writepos = buf;
	  for(x = xmin; x < xmax; x++) {
	    render_pixel_imagebuffer(point, plane, &red, &green, &blue);
	    *writepos++ = red;
	    *writepos++ = green;
	    *writepos++ = blue;
	    point++;
	  }
	  if( !write_pixels_ppmwriter(p, buf, xmax - xmin) ) {
	    retval = FALSE;
	    break;
	  }
	}
	free(buf);
      }
    }
    return retval && consistency_check_ppmwriter(p);
  }
  return FALSE;
}


bool_t fillhist_imagebuffer(imagebuffer_t *ib, int plane) {
  int x, y;
  imagepoint_t *point;

  if( ib ) {
    for(y = 0; y < ib->height; y++) {
      point = ib->image + y * ib->width; 
      for(x = 0; x < ib->width; x++) {
	WRITEBW(point->pixl,plane,(point->pixf > 0));
	point++;
      }
    }
    return TRUE;
  }
  return FALSE;
}


bool_t simplecrop_bw_imagebuffer(imagebuffer_t *ib,
				 int xmin, int ymin, int xmax, int ymax,
				 int plane) {
  int x, y;
  imagepoint_t *point;

  if( ib ) {
    ib->crop.xmin = ib->width;
    ib->crop.xmax = 0;
    ib->crop.ymin = ib->height;
    ib->crop.ymax = 0;
    ib->crop.coverage = 0;

    for(y = ymin; y < ymax; y++) {
      point = ib->image + y * ib->width + xmin; 
      for(x = xmin; x < xmax; x++) {
	if( READBW(point->pixl,plane) ) {
	  ib->crop.xmin = MIN(ib->crop.xmin, x);
	  ib->crop.xmax = MAX(ib->crop.xmax, x);
	  ib->crop.ymin = MIN(ib->crop.ymin, y);
	  ib->crop.ymax = MAX(ib->crop.ymax, y);
	  ib->crop.coverage++;
	}
	point++;
      }
    }
    return TRUE;
  }
  return FALSE;
}

/* this function creates a crop mask in plane sp. The 
 * crop mask is the connected component of the support
 * of the pixel histogram, containing a number of seed
 * values. The seed values are obtained by computing a
 * level set of the histogram, so level must be <= 100.
 *
 * The function destroys the contents of dp.
 */
bool_t crop_bw_imagebuffer(imagebuffer_t *ib, int level, int dp, int sp) {
  int x, y;
  imagepoint_t *point;
  mask_t mask;
  if( ib ) {

    /* make sure pixl0 contains histogram support,
     * and clear pixl1 
     */
    for(y = 0; y < ib->height; y++) {
      point = ib->image + y * ib->width; 
      for(x = 0; x < ib->width; x++) {
	WRITEBW(point->pixl,dp,(point->pixf > 0));
	CLEARBW(point->pixl,sp);
	point++;
      }
    }

    ib->crop.xmin = ib->width;
    ib->crop.xmax = 0;
    ib->crop.ymin = ib->height;
    ib->crop.ymax = 0;

    if( create_mask(&mask) ) {
      /* this is more efficient than it looks */
      for(y = 0; y < ib->height; y++) {
	point = ib->image + y * ib->width; 
	for(x = 0; x < ib->width; x++) {
	  if( (point->pixf > level) && (READBW(point->pixl,sp) == 0) ) {
	    make_component_bw_mask(&mask, ib, x, y, dp, sp);
	    ib->crop.xmin = MIN(ib->crop.xmin,mask.c.xmin);
	    ib->crop.xmax = MAX(ib->crop.xmax,mask.c.xmax);
	    ib->crop.ymin = MIN(ib->crop.ymin,mask.c.ymin);
	    ib->crop.ymax = MAX(ib->crop.ymax,mask.c.ymax);
	  } 
	  point++;
	}
      }
      free_mask(&mask);
    }

    /* compute coverage */
    ib->crop.coverage = 0;
    for(y = 0; y < ib->height; y++) {
      point = ib->image + y * ib->width; 
      for(x = 0; x < ib->width; x++) {
	if( READBW(point->pixl,sp) ) {
	  ib->crop.coverage += point->pixf;
	}
	point++;
      }
    }

    return TRUE;
  }
  return FALSE;
}


bool_t box_bw_imagebuffer(imagebuffer_t *ib,
			   int xmin, int ymin,
			   int xmax, int ymax,
			   int plane, int val) {
  int x, y;
  imagepoint_t *point;

  if( ib ) {
    xmin = MAX(xmin,0);
    ymin = MAX(ymin,0);
    xmax = MIN(xmax,ib->width);
    ymax = MIN(ymax,ib->height);

    point = ib->image + ymin * ib->width + xmin; 
    for(x = xmin; x < xmax; x++) {
      WRITEBW(point->pixl,plane,val);
      point++;
    }
    
    for(y = ymin + 1; y < ymax; y++) {
      point = ib->image + y * ib->width + xmin; 
      WRITEBW(point->pixl,plane,val);
      for(x = xmin; x < xmax; x++) {
	point++;
      }
      WRITEBW(point->pixl,plane,val);
    }

    if( ymax < ib->height ) {
      point = ib->image + ymax * ib->width + xmin; 
      for(x = xmin; x < xmax; x++) {
	WRITEBW(point->pixl,plane,val);
	point++;
      }
    }

    return TRUE;
  }
  return FALSE;
}

bool_t fill_bw_imagebuffer(imagebuffer_t *ib,
			   int xmin, int ymin,
			   int xmax, int ymax,
			   int plane, int val) {
  int x, y;
  imagepoint_t *point;

  if( ib ) {
    xmin = MAX(xmin,0);
    ymin = MAX(ymin,0);
    xmax = MIN(xmax,ib->width);
    ymax = MIN(ymax,ib->height);
    for(y = ymin; y < ymax; y++) {
      point = ib->image + y * ib->width + xmin; 
      for(x = xmin; x < xmax; x++) {
	WRITEBW(point->pixl,plane,val);
	point++;
      }
    }
    return TRUE;
  }
  return FALSE;
}

/* move a piece of "from" plane over to "to" plane, and clear "from",
 * but don't clear "to".
 */
bool_t mv_bw_imagebuffer(imagebuffer_t *ib, 
			 int xmin, int ymin, int xmax, int ymax,
			 int from, int to) {
  int x, y;
  imagepoint_t *point;

  if( ib ) {
    for(y = ymin; y < ymax; y++) {
      point = ib->image + y * ib->width; 
      for(x = xmin; x < xmax; x++) {
	if( READBW(point->pixl, from) ) {
	  SETBW(point->pixl,to);
	}
	CLEARBW(point->pixl,from);
	point++;
      }
    }
    return TRUE;
  }
  return FALSE;
}

