/* 
 * Copyright (C) 2008 Laird Breyer
 *  
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * Author:   Laird Breyer <laird@lbreyer.com>
 */

/* 
 * Note to regenerate the Makefile and configure script:
 * aclocal
 * autoconf
 * touch NEWS README AUTHORS THANKS ChangeLog
 * automake -a
 *
 * You can also just run the bootstrap script.
 */

#include "util.h"
#include "filemagic.h"
#include "pbmreader.h"
#include "boxreader.h"
#include "boxwriter.h"
#include "ppmwriter.h"
#include "imagebuffer.h"
#include "boxlist.h"
#include "boxeditor.h"
#include "tessboxes.h" /* make sure this is last */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <stdlib.h>
#include <string.h>
#include <ctype.h>

/* global variables */
extern char *progname;
extern char *inputfile;
extern long inputline;

extern int cmd;
extern options_t u_options;
extern options_t m_options;

#ifndef EXIT_SUCCESS
#define EXIT_SUCCESS 0
#define EXIT_FAILURE 1
#endif

int exit_code = EXIT_SUCCESS;
extern linesize_t default_linesize;
extern pagesize_t default_pagesize;
extern linesize_t linesize;
extern pagesize_t pagesize;
extern size_t default_bufalign;
extern size_t bufalign;
extern size_t default_path_max;
extern size_t path_max;

imagetype_t outputformat = tP4;

/* neighbourhood size for features */
int nsize = 16;
/* multiplier for feature weights (ie Boltzmann factor, inverse temp, etc) */
float sharpness = 1.0;

char *inputimage;
char *boxfile;

char *goto_string = NULL;
long int goto_x = -1;
long int goto_y = -1;

/* when editing, shift the highlighted box off center for usability */
double box_shift = 0.5;

/***********************************************************
 * MISCELLANEOUS FUNCTIONS                                 *
 ***********************************************************/

static void usage(/*@unused@*/ char **argv) {
  fprintf(stderr, 
	  "\n");
  fprintf(stderr, 
	  "%s PBMIMAGE BOXFILE\n", progname);
  fprintf(stderr, 
	  "\n");
  fprintf(stderr, 
	  "      outputs a PPM file containing bounding boxes overlaid on the input image.\n");
  fprintf(stderr, 
	  "\n");
  fprintf(stderr, 
	  "%s -e PBMIMAGE BOXFILE [-g X,Y]\n", progname);
  fprintf(stderr, 
	  "\n");
  fprintf(stderr, 
	  "      edits the BOXFILE interactively (and saves the new version)\n");
  fprintf(stderr, 
	  "\n");
  fprintf(stderr, 
	  "%s -V\n", progname);
  fprintf(stderr, 
	  "\n");
  fprintf(stderr, 
	  "      prints program version.\n");
}


/***********************************************************
 * OPTION PROCESSING                                       *
 ***********************************************************/

int set_option(int op, char *optarg) {
  int c = 0;
  switch(op) {
  case 'e':
    u_options |= (FL_EDIT);
    break;
  case 'g':
    if( optarg ) {
      u_options |= (FL_EDIT|FL_GOTO);
      goto_string = optarg;
      while(*optarg) {
	if( !isdigit(*optarg) ) {
	  *optarg = ' ';
	}
	optarg++;
      }
    } 
    break;
  case 's':
    if( optarg && *optarg ) {
      u_options |= FL_BOXSHIFT;
      box_shift = strtod(optarg, NULL);
      if( box_shift < 0.01 || box_shift > 0.99 ) {
	errormsg(E_FATAL, "box shift is not in range [0.01, 0.99]\n");
      }
    } 
    break;
  case 'V':
    fprintf(stdout, "%s version %s\n", progname, VERSION);
    fprintf(stdout, COPYBLURB, progname);
    fprintf(stdout, "\n");
    exit(EXIT_SUCCESS);
    break;
  default:
    c--;
    break;
  }
  return c;
}

void sanitize_options() {
  char *q;
  /* consistency checks */
  if( u_options & FL_GOTO ) {
    q = goto_string;
    goto_x = strtol(q, &q, 10);
    goto_y = strtol(q, &q, 10);
  }
}


void init_file_handling() {
}

void cleanup_file_handling() {
}

bool_t process_file(const char *filename, imagebuffer_t *ib, 
		    boxlist_t *bl, int pass) {
  int fd;
  pbmreader_t p;
  boxreader_t b;

  fd = open(filename, O_RDONLY);
  if( fd != -1 ) {
    switch(pass) {
    case 0:
      switch(identify_file(fd)) {
      case tPBM:
      case tGZIP:
	/* read in an image, and fill the statistics */
	if( create_pbmreader(&p, fd) ) {
	  if( !grow_imagebuffer(ib, p.width, p.height) ) {
	    errormsg(E_FATAL, "memory allocation error.\n");
	  }
	  if( !fillimage_pixl0_pbmreader(&p, ib) ) {
	    errormsg(E_WARNING, 
		     "the file %s may be corrupt.\n", filename);
	  }
	  free_pbmreader(&p);
	}
	break;
      default:
      case tUNKNOWN:
	errormsg(E_FATAL, 
		 "unrecognized file format %s (pass = %d).\n", 
		 filename, pass);
      }
      break;
    case 1:
      /* read the box coordinates and write the modified image */
      if( create_boxreader(&b, fd) ) {
	if( !parse_boxreader(&b, ib, bl) ) {
	  errormsg(E_WARNING, 
		   "the file %s may be corrupt.\n", filename);
	}
	free_boxreader(&b);
      }
      break;
    default:
      errormsg(E_FATAL, 
	       "unrecognized file format %s (pass = %d).\n", 
	       filename, pass);
      break;
    }
    close(fd);
    return TRUE;
  }
  return FALSE;
}

int post_process(imagebuffer_t *ib, boxlist_t *bl, int pass) {
  ppmwriter_t p;
  boxeditor_t e;
  boxwriter_t w;
  int fd;
  int i = 0;
  char *s;

  if( ib ) {
    switch(pass) {
    case 0:
      fd = fileno(stdout);
      if( create_ppmwriter(&p, fd, tP6) ) {
	save_color_imagebuffer(ib, &p, tP6, 0);
	free_ppmwriter(&p);
      }
      break;
    case 1:
      if( create_boxeditor(&e, ib, bl) ) {
	if( u_options & FL_GOTO ) {
	  if( gotoxy_boxeditor(&e, goto_x, goto_y) ) {
	    i = run_boxeditor(&e);
	  } else {
	    i = FALSE;
	    fprintf(stderr, "Unknown coordinates (%ld,%ld). ",
		    goto_x, goto_y);
	  }
	} else {
	  i = run_boxeditor(&e);
	}
	free_boxeditor(&e);
      }
      if( i == FALSE ) {
	fprintf(stderr, "Editor canceled. Box file not modified.\n");
      } else {
	s = malloc(strlen(boxfile) + 2);
	if( s ) {
	  strcpy(s, boxfile);
	  strcat(s, "~");
	  rename(boxfile, s);

	  fd = open(boxfile, O_WRONLY|O_CREAT, S_IRUSR|S_IWUSR);
	  if( create_boxwriter(&w, fd) ) {
	    for(i = 0; i < bl->count; i++) {
	      if( !write_boxwriter(&w, &bl->items[i]) ) {
		i = -1;
		break;
	      }
	    }
	    free_boxwriter(&w);
	  }
	  close(fd);

	  if( i == -1 ) {
	    unlink(boxfile);
	    rename(s, boxfile);
	    fprintf(stderr, "An error occurred. Box file not modified.\n");
	    return FALSE;
	  } else {
	    fprintf(stderr, "Saving box file %s\n", boxfile);
	  }
	  free(s);
	}
      }
      break;
    default:
      return FALSE;
    }
  }
  return TRUE;
}

int main(int argc, char **argv) {

  signed char op;
  imagebuffer_t ib;
  boxlist_t bl;
  int post_cmd;

  progname = "tessboxes";
  inputfile = "";
  inputline = 0;

  init_signal_handling();

  /* defaults */
  u_options = (FL_BB);

  /* parse the options */
  while( (op = getopt(argc, argv, 
		      "eg:s:V")) > -1 ) {
    if( op == 'e' ) {
      init_boxeditor(&argc, &argv);
    }
    set_option(op, optarg);
  }

  /* end option processing */
  sanitize_options();

  inputimage = argv[optind++];
  if( !inputimage ) {
    errormsg(E_ERROR, "missing input file (in PBM format).\n");
    usage(argv);
    exit(EXIT_FAILURE);
  }
  boxfile = argv[optind];
  if( !boxfile ) {
    errormsg(E_ERROR, "missing boxfile (in tesseract format).\n");
    usage(argv);
    exit(EXIT_FAILURE);
  }

  if( create_imagebuffer(&ib) ) {
    if( create_boxlist(&bl) ) {
      init_file_handling();

      if( !process_file(inputimage, &ib, &bl, 0) ) {
	errormsg(E_FATAL, "bad input image, is it PBM format?\n");
      }

      if( !process_file(boxfile, &ib, &bl, 1) ) {
	errormsg(E_FATAL, "bad box file, is it tesseract format?\n");
      }

      if(cmd & FL_QUITNOW) {
	exit_code = EXIT_FAILURE;
      } else {
	post_cmd = (u_options & FL_EDIT) ? 1 : 0;
	if( !post_process(&ib, &bl, post_cmd) ) {
	  errormsg(E_FATAL, "unrecoverable error, output may be unusable\n");
	}
      }

      cleanup_file_handling();

      free_boxlist(&bl);
    }

    free_imagebuffer(&ib);
  } else {
    errormsg(E_FATAL, "cannot create imagebuffer.\n");
  }
  cleanup_signal_handling();

  return exit_code;
}


